"""
This program was written jointly by Vasily Dolgushev and Geoffrey Schneider

The goal of this program is to define
the elementary insertions for vectors in BT.


Trees are represented nested lists of edges.
The edges are ordered using the planar structure.
They appear in this order.

ABOUT BRACE TREES:

We represent a brace (planar) tree T by the nested list of
its edges. The edges show up in the order compatible with the
planar strutcure. We assume that trees "GROW UP". In other words,
if we have [1,2] (resp. [2,1]) then 1 (resp. 2) is closer to the root.

A brace tree is called standard if its vertices are ordered
according to the planar structure. For example, [[1,2],[2,3]] is
a standard brace tree and  [[2,1],[1,3]] is not.

"""




# coding: utf-8
#import itertools
from itertools import combinations_with_replacement as cr
#from itertools import permutations as perm

from copy import deepcopy as deepcopy

#import time
#import math



from LinCombGraphs import Sign as Sign
from LinCombGraphs import mult as mult
from LinCombGraphs import Simplify as Simplify
from LinCombGraphs import linExt as linExt
from LinCombGraphs import Val as Val
from LinCombGraphs import NumIn as NumIn
from LinCombGraphs import Vert as Vert
from LinCombGraphs import NumVert as NumVert

from TwBT import H2Comp as H2Comp
from TwBT import Comp2Hlc as Comp2Hlc


"""
FIRST, WE DEFINE VARIOUS ELEMENTARY FUNCTIONS
"""


"""
Above gives the tuple of vertices in a brace tree T which are
adjacent to i and above i. The vertices in this list
are ordered according to their planar structure.
"""

#Tested
def Above(T,i):
    return tuple(ed[1] for ed in T if ed[0] == i)


"""
EdgesFrom(T,i) returns the list of edges which are adjacent to i and originate from i.
The list is ordered according to the planar structure.
"""
#Tested
def EdgesFrom(T,i):
    return [ed for ed in T if ed[0] == i]

"""
RightMost returns the last edge in EdgesFrom(T,i)
NOTE: there should be edges originating from i!
"""
#Tested
def RightMost(T,i):
    return EdgesFrom(T,i)[len(EdgesFrom(T,i))-1]


"""
LeftMost returns the first edge in EdgesFrom(T,i)
NOTE: there should be edges originating from i!
"""
#Tested
def LeftMost(T,i):
    return EdgesFrom(T,i)[0]


"""
EdgeBelow returns the edge which is adjacent to i and below i
We assume that i IS NOT the lowest vertex.
"""
#Tested
def EdgeBelow(T,i):
    for ed in T:
        if ed[1] == i:
            return ed



"""
A function sig acts on a graph G (which could be a tree).
Of course, sig should be defined on all vertices.
If G is a standard brace tree, sig(G) is usually not standard.
For example, Act((2,1,3),[[1,2],[2,3]]) returns [[2,1],[1,3]]

NOTE: sig is a tuple or a list
"""

#Tested
def Act(sig, G):
    return [ [sig[ed[0]-1], sig[ed[1]-1]] for ed in G]



#Tested
def actInv(sig, TT):
    """
    Returns the tree obtained by acting by the inverse of sig on TT.
    """
    inv = tuple(sig.index(k)+1 for k in range(1, len(sig)+1))
    return Act(inv, TT)

"""
We typically need the shift function.
The function shift(i,r, BN) produces a tuple of the form
(1,2,...i-1, i+r, i+r+1,..., BN+r).
We think of this tuple as the function
1->1, 2 -> 2, ... i-1 -> i-1, i -> i+r, and so on.
"""
#Tested
def shift(i,r,BN):
    out = tuple(j for j in range(1, i))
    adding = tuple(j+r for j in range(i, BN+1))
    return out + adding



"""
shiftT(i,r,G) shifts the labels on vertices of G up by r starting with vertex i.
"""
#Tested
def shiftT(i,r,G):
    fun = shift(i,r,NumVert(G))
    return Act(fun,G)


"""
TopREdge(TT, i) returns the last edge of the subtree which originates
from vertex i of TT. We assume that i has val > 1. Othewise,
TopREdge does NOT work.
"""
#Tested
def TopREdge(TT, i):
    ed = RightMost(TT,i)
    while True:
        if Val(TT,ed[1]) != 1:
            ed = RightMost(TT,ed[1])
        else:
            break
    return ed

"""
subTree returns the subtree with the lowest vertex i.
If i is a leaf then the output is the empty list.
"""
#Tested
def subTree(TT, i):
    if TT[0][0] == i:
        return TT
    if Val(TT,i) == 1:
        return []
    else:
        return TT[ TT.index(LeftMost(TT,i)) : TT.index(TopREdge(TT, i))+1 ]




"""
We are "falling down" along right hand side of
the brach starting with vertex j.

fall(TT,j) returns the tuple of vertices starting with j
then the one adjacent to j (and below j) and then down till
the vertex where we have a "braching to the right".
This vertex with braching is the last of the tuple.

For example, if T = [[1,2],[2,3],[1,4]] then
fall(T,3) = (3,2,1), fall(T,4) = (4,1)


NOTE: fall will NOT WORK if j is the lowest vertex of TT.
"""

#Tested
def fall(TT, j):
    out = (j,)
    aux = [EdgeBelow(TT,j)]
    while True:
        ed = aux[len(aux)-1]
        if ed[0] == TT[0][0]:
            out = out + (ed[0],)
            break
        if ed == RightMost(TT, ed[0]):
            out = out + (ed[0],)
            aux = aux + [EdgeBelow(TT,ed[0])]
        else:
            out = out + (ed[0],)
            break

    return out


#Tested
def secTree(TT):
    """
    Returns the list of places of insertion with the associated vertex and edges.
    places of insertions and vertices are given as a tuple (place of insertion, associated vertex). e.g.
    secTree([[1,2]]) returns [(1,1),[1,2],(2,2),(3,1)].
    """
    intr = [(1,TT[0][0])]
    current = 2
    for ed in TT:
        intr.append(ed)
        if Val(TT, ed[1]) == 1:
            f = fall(TT, ed[1])
            intr.extend([(k+current,f[k]) for k in range(len(f))])
            current += len(f)
        else:
            intr.append((current, ed[1]))
            current += 1
    return intr



"""
edgesLeft returns the list of edges of TT which
precede the EdgeBelow(TT,i) together with EdgeBelow(TT,i).
"""

#Tested
def edgesLeft(TT, i):
    if i == TT[0][0]:
        return []
    else:
        ed = EdgeBelow(TT,i)
        return TT[0:TT.index(ed)+1]


"""
edgesRight returns the list of edges of TT which
which supersede TopREdge(TT, i), i.e. all edges
after the edges of subTree(TT,i)
"""

#Tested
def edgesRight(TT, i):
    if i == TT[0][0]:
        return []
    if i == Vert(TT)[len(Vert(TT))-1]:
        return []
    if Val(TT,i) == 1:
        return TT[ TT.index(EdgeBelow(TT,i))+1 : len(TT) ]
    else:
        return TT[ TT.index(TopREdge(TT,i)) +1 : len(TT) ]

"""
i is a vertex of TT. prune(TT, i) returns
the list of branches originating from i
(in the usual planar order).

If NO edges originate from i then prune returns []
"""
#Tested
def prune(TT, i):
    if NumIn(TT,i) == 0:
        return []
    else:
        return [ [[i,j]] + subTree(TT,j) for j in Above(TT,i)]




"""
graft(...) grafts trees in Tlist onto TT at the positions in the tuple sec.
sec may have repetitions!!! len(sec) must equal the length of TList
The output of graft is a brace tree

NOTE: If el[0] is not in sec then the second loop
has zero iterations.

At the end, the list Aux must be empty.

[ [el[1],B[0][1]] ] + B[1:] is the modification of the brach B
B[0][0] should be replaced by el[1]

NOTE: For every brach B the lowest vertex B[0][0] is necessarily
univalent!

NO shifts for vertices here!
"""

#Tested
def graft(TT, Tlist, sec):
    Out = []
    Aux = deepcopy(Tlist)
    for el in secTree(TT):
        if type(el) == list:
            Out.append(el[:])
        else:
            for a in range(sec.count(el[0])):
                B = Aux.pop(0)
                Out = Out + [ [el[1], B[0][1]] ] + B[1:]

    return Out


#Tested
def treeDeg(TT):
    return 1-NumVert(TT)

#Tested
def blackCan(TT):
    """
    Puts black vertices in canonical order.
    """
    order = []
    for ed in TT[1]:
        for v in ed:
            if v <= TT[0] and (not v in order): order.append(v)
    order = tuple(order + [k for k in range(TT[0]+1,NumVert(TT[1])+1)])
    return [TT[0], actInv(order, TT[1])]

#Tested using hCirc
def twCirc(T1,i,T2):
    """
    Composes two elements of TwBT, represented as lists with two elements, the
    0th giving r, the number of black vertices, and the 1st a list of edges,
    with the first r representing the black vertices. Black vertices must be
    labelled in canonical order, and will be returned in canonical order.
    """
    N2tot = NumVert(T2[1])
    sh2fnc = tuple([k for k in range(1,T2[0]+1)]+[k for k in range(T1[0]+T2[0]+i,T1[0]+N2tot+i)])
    T2sh = Act(sh2fnc,T2[1])
    j = T2sh[0][0]
    sh1fnc = tuple([k for k in range(T2[0]+1,T1[0]+T2[0]+1)]+[k for k in range(T1[0]+T2[0]+1,T1[0]+T2[0]+i)]+[j]+[k for k in range(T1[0]+N2tot+i,NumVert(T1[1])+N2tot)])
    T1sh = Act(sh1fnc,T1[1])
    N2 = N2tot
    V1 = Vert(T1sh)
    V2 = Vert(T2sh)
    if NumIn(T1sh, j) == 0:
        Aux = edgesLeft(T1sh,j) + T2sh + edgesRight(T1sh,j)
        return [[Sign(T1sh + T2sh, Aux), blackCan([T1[0]+T2[0],Aux])]]
    else:
        Blist = prune(T1sh,j)
        graftList = [graft(T2sh, Blist, seq) for seq in cr(range(1,2*N2),len(Blist))]
        Out = []
        for T in graftList:
            Aux = edgesLeft(T1sh,j) + T + edgesRight(T1sh,j)
            Nd = []
            for ed in Aux:
                if ed[0] in V2 and ed[1] in V1:
                    Nd.append([j,ed[1]])
                else:
                    Nd.append(ed[:])
                # we need the list Nd for the right sign factor
            Out.append([Sign(T1sh + T2sh, Nd), blackCan([T1[0]+T2[0],Aux[:]])])
        return Simplify(Out)

#Tested
def whiteV(TT, j):
    """
    Takes TT in TwBT with n white vertices and r black vertices, and returns the
    tree obtained by replacing j <= r by a white vertex with label n+1.
    """
    NR = NumVert(TT[1])
    fnc = tuple([k for k in range(1,j)] + [NR] + [k-1 for k in range(j+1,NR+1)])
    return [TT[0]-1, Act(fnc,TT[1])]

#Tested
def twDeg(TT):
    """
    Returns the degree of TT in TwBT
    """
    return 2*TT[0] + treeDeg(TT[1])

#
# The three c-presented brace trees are for the function twDiff   
Tbb = [2,[[1,2]]]
T1b = [1,[[2,1]]]
Tb1 = [1,[[1,2]]]
# Tested using hDiff
def twDiff(TT):
    """
    Returns the differential of TT in TwBT.
    """
    Out = []
    N = NumVert(TT[1]) - TT[0]
    for v in range(1,TT[0]+1):
        Out.extend(mult((twDeg(TT)%2)*2 - 1,twCirc(whiteV(TT,v),N+1,Tbb)))
    Out.extend(twCirc(T1b,1,TT))
    Out.extend(twCirc(Tb1,1,TT))
    for i in range(1,N+1):
        Out.extend(mult((twDeg(TT)%2)*2 - 1,twCirc(TT,i,T1b)))
        Out.extend(mult((twDeg(TT)%2)*2 - 1,twCirc(TT,i,Tb1)))
    return Simplify(Out)

"""
hCirc computes the insertion into the i-th slot and it uses
the h-presentation of brace trees! The output is linear combination 
of h-presented trees. 
"""
#Tested
def hCirc(T1,i,T2):
    TT1 = H2Comp(T1)
    TT2 = H2Comp(T2)
    return Comp2Hlc(twCirc(TT1,i, TT2))


    
"""
hDiff compute the differential of an h-presented tree.
"""
#Tested
def hDiff(T):
    TT = H2Comp(T)
    return Comp2Hlc(twDiff(TT))
    
"""
hDifflc is the extension of hDiff to linear combinations
"""
#Tested
def hDifflc(x):
    return linExt(hDiff, x)


# Examples of brace trees in different formats FOR TESTING:


t1 = [4,[[5,1],[1,7],[7,2],[1,3],[5,6],[6,8],[6,4]]]
t2 = [2,[[1,3],[3,2],[1,4]]]

Tbr = [0,[[1,2]]]
Tbrr = [0,[[2,1]]]

Del = [0,[[1,2],[1,3]]]

        

hCup = [[(1,),1], [(1,),2]]
Cup = [1, [[1, 2], [1, 3]]]
         
hCupOpp = [[(1,),2], [(1,),1]]
CupOpp = [1, [[1, 3], [1, 2]]]
            

hCup3 = [[(1,),1], [(1,),2], [(1,),3]]
Cup3 = [1, [[1, 2], [1, 3], [1, 4]]]

          
hCCup = [[(1,),(2,)], [(2,),1], [(2,),2], [(1,),3]]
CCup = [2, [[1, 2], [2, 3], [2, 4], [1, 5]]]

          
hCupC = [[(1,),1],[(1,),(2,)], [(2,),2], [(2,),3]]
CupC = [2, [[1, 3], [1, 2], [2, 4], [2, 5]]]
          
      
T12 = [[1,2]]
T21 = [[2,1]]

T123 = [[1,2],[1,3]]
L123 = [[1,2],[2,3]]

TT = [[1,2],[2,3],[1,4],[4,5],[4,6] ]

Ts = [[3, 5], [5, 2], [3, 1], [1, 4], [1, 6]]

Tss = [ [2,(1,)], [(1,),(2,)], [(2,),3],[(2,),4], [(1,),1]]

"""
Comna is the differential of T123:
"""
Comna = [[-1, [[1, (1,)], [(1,), 2], [(1,), 3]]],
 [-1, [[(1,), 2], [(1,), 3], [(1,), 1]]],
 [1, [[(1,), 2], [(1,), 1], [1, 3]]],
 [1, [[(1,), 2], [(1,), 1], [(1,), 3]]],
 [-1, [[(1,), 1], [1, 2], [(1,), 3]]],
 [-1, [[(1,), 1], [(1,), 2], [(1,), 3]]]]


"""
To "unpinckle" WE DO THIS:

FO = open('BT12','rb')  # Now we can read the file storage
BTback = pickle.load(FO) #

"""

