"""
This module defines operations performed in Conv(Ger*,Br) viewed as
\prod_n (Br(n) \otimes Ger(n))_{S_n}. Vectors in Br are given as
linear combinations of h-presented trees.
An element is given as a
list whose elements are lists whose first entry is a coefficient, and second
entry is the element of the basis given as a tuple where the first element is a
brace tree and the second a Ger word.
"""

import pickle


import itertools
from itertools import combinations_with_replacement as cr
from itertools import permutations as perm
from itertools import product as prod

import time
import math


from copy import deepcopy as deepcopy
#NOTE: deepcopy can be used for sympy objects!
from sympy import *

# for a nested list L, Matrix(L) produces the corresponding sympy matrix
# for a sympy matrix m, the command L = m.tolist() returns the corresponding nested list.

from LinCombGraphs import inv as inv
from LinCombGraphs import mult as mult
from LinCombGraphs import Simplify as Simplify
from LinCombGraphs import toVect as toVect
from LinCombGraphs import toLC as toLC
from LinCombGraphs import Terms as Terms
from LinCombGraphs import linExt as linExt
from LinCombGraphs import bilinExt as bilinExt


from LinCombGraphs import toCol as toCol
from LinCombGraphs import Null as Null
from LinCombGraphs import Solve as Solve
from LinCombGraphs import NZlist as NZlist

from Bases import loadConvCirc as loadConvCirc
from Bases import loadConvBul as loadConvBul

from BTCirc import hCirc as hCirc
from BTCirc import hDiff as hDiff
from TwBT import arity as arity
from TwBT import ActBr as ActBr
from TwBT import getPermBr as getPermBr

from Ger import InsGG as InsGG
from Ger import degG as degG
from Ger import ActG as ActG
from Ger import toStanGer as toStanGer


"""
System is class which is reserved for the a collection of vectors.
Col is the collection of vectors. size is the length of Col.
self.terms is the tuple of all terms of vectors in Col without 
repetition. 

self.Mtx() returns the matrix of Col with respect to the basis self.terms
The output of self.Mtx is the list [[col1], [col2], ...]

LinIndep(self) returns True if the collection Col is linearly independent.
Otherwise False

For example, if B = System(tuple of vectors)
Then B.LinIndep() is True if the tuple of vectors is linearly indepedent. 

InSpan(self, u) returns True if the vector u is in the span of Col.
NOTE: It is assumed that the linear combination u is simplified.

"""

#Tested
class System:
    def __init__(self, Col):
        self.Col = Col
        self.size = len(Col)
        self.terms = ()
        for v in self.Col:
            for t in Terms(v):
                if t not in self.terms:
                    self.terms = self.terms + (t,)
                    
    def Mtx(self):
        L = []
        for w in self.Col:
            v = toVect(w,self.terms)
            L.append(deepcopy(v))
        return L
    
    def LinIndep(self):
        M = Matrix(self.Mtx())
        M = M.T
        piv = M.rref()[1]
        return len(piv) == self.size
     
    def InSpan(self, u):
        for t in Terms(u):
            if t not in self.terms:
                return False
        L = self.Mtx()
        v = toVect(u,self.terms)
        L.append(deepcopy(v))
        M=Matrix(L)
        M = M.T
        piv = M.rref()[1]
        return self.size not in piv
        


        
"""     
# These vectors were used for testing the class System:

v = [[5, (1,)], [-2, (2,)], [1,(3,)]]
    
vv = [[11, (2,)], [7,(3,)]]

w = [[11, (5,)]]
""" 

"""
L is a linear combination of tensor monomials in Br \otimes Ger
toStandConv(L) returns the corresponding linear 
combination the (h-presented) trees which have the standard form. 
Note that it does not simplify!
"""
#Tested
def toStandConv(L):
    out =[]
    for x in L:
        (T, w) = x[1]
        s = inv(getPermBr(T))
        ww = toStanGer(ActG(s,w)) # this is a lin comb of Ger words
        TT = ActBr(s, T)
        for a in ww:
            out.append(deepcopy([x[0]*a[0], (TT, a[1]) ]))

    return out


"""
dConv(x) returns the differential of the tensor monomial in Br \otimes Ger.
The output is a linear combination of tensor monomials. 
We apply toStandConv but we do not simplify!
"""
#Tested
def dConv(x):
    dx0, out = hDiff(x[0]), []
    for a in dx0:
        out.append([a[0],(a[1],x[1])])
        
    return toStandConv(out)
    
"""
dConvlc is the linear extension of dConv. 
The function Simplify is USED in linExt
"""
#Tested
def dConvlc(x):
    return linExt(dConv,x)

"""
CircConv(x,i,y) returns the linear combination 
corresponding to x[0] \circ_{i} y[0] \otimes x[1] \circ_{i} y[1] 
with the appropriate sign factor. The brace trees x[0] and 
y[0] are h-presented. 

The function also brings the brace trees to the standard form. 
But no simplification...
"""
#Tested
def CircConv(x,i,y):
    out = []
    s = (-1)**(degG(x[1])*len(y[0])) # this is the sign.
    L = hCirc(x[0], i, y[0])
    LL = InsGG(x[1], i,y[1])
    for a in L:
        for b in LL:
            out.append([s*a[0]*b[0], (a[1],b[1])])
    
    return toStandConv(out)
    
"""
pLie returns the pre-Lie product of two tensor monomials. 
The output is a linear combination of tensor monomials.
The function Simplify IS NOT used!
"""
#Tested
def pLie(x,y):
    out, n = [], arity(x[0])
    for i in range(1, n+1):
        out = out + CircConv(x,i,y)
        
    return out


"""
pLielc is the bilinear extension of pLie. 
The function Simplify is "hidden in" bilinExt 
"""
#Tested
def pLielc(x,y):
    return bilinExt(pLie,x,y)
    
"""
lieConv(x,y) returns the Lie bracket of two tensor monomials in 
Br \otimes \Ger
Note that Simplify is NOT used. 
"""
#Tested
def lieConv(x,y):
    s = -(-1)**((len(x[0])+degG(x[1]))*(len(y[0])+degG(y[1])))
    return pLie(x,y) + mult(s,pLie(y,x))
 
"""
lieConvlc is the bilinear extension of the bracket lieConv. 
Recall that Simplify is "hidden in" bilinExt.
"""
#Tested
def lieConvlc(x,y):
    return bilinExt(lieConv, x, y)

"""
Curv returns the modified Curvature of degree 1 elements in Br \otimes Ger. 
Here N is the denominator. In other words, the MC sprout is (1/N)*a
Note that "Simplify" is USED.
"""

# Tested. Use with care! For arity > 3, the function produces 
# a huge number of terms!
def Curv(N, a):
    out = mult(N,dConvlc(a)) + pLielc(a,a) # Just look at this beautiful Maurer-Cartan equation!!! :-)
    return Simplify(out)
    
"""
The stuff below is for testing:
"""

"""
minArity returns the minimal arity of terms in a linear combination
L \in Br \otimes \Ger. It is used mostly for testing. 
"""
#Tested
def minArity(L):
    ar = tuple(arity(t[1][0]) for t in L)
    return min(ar)

"""
CollectMinArity(L) returns the truncation of L. 
This truncation is obtained by discarding terms of arity > 
than the minimal one. 
"""
#Tested
def CollectMinArity(L):
    m = minArity(L)
    return [t for t in L if arity(t[1][0]) == m]


# Trying the find an extendable 3rd MC sprout.
# Arities 2,3,4.

DM =240 # This is the basic denominator.
            
    
"""
alnc (alnb) is the arity n part without neutral vertices 
(resp. with 1 neutral vertex)
"""    

al2c = [ [240,([[1,2]], [(1,),(2,)])] ]

al2b = [ [120,([[(1,),1], [(1,),2]], [(1,2)])] ]
          
al2 = al2b + al2c

"""
This piece of code can be used to find arity 3 part: 

CB = loadConvBul(); CC = loadConvCirc()
C3 = CB[2] + CC[2]
#del CB; del CC # to go easy on the memory...

Tup3 = tuple(mult(DM, dConvlc([[1,v]])) for v in C3) 

RHS3  = mult(-1, pLielc(al2, al2))

Sys3 = System(Tup3 +(RHS3,)) 

Aug3 = Matrix(Sys3.Mtx()); Aug3 = Aug3.T

Res3 = Solve(Aug3) # Solving the linear system -DM* d(al3b) =  ...

# Res3b[0] is True. So the system is consistent.  

al3new = toLC(Res3[1][:],C3)

END of the piece for finding arity 3 part. 
"""       
al3b = [[-80, ([[(1,), 1], [(1,), 2], [(1,), 3]], [(1, 2, 3)])],
 [-40, ([[(1,), 1], [(1,), 2], [(1,), 3]], [(2, 1, 3)])],
 [-20, ([[(1,), 1], [1, 2], [(1,), 3]], [(1, 2, 3)])],
 [-20, ([[(1,), 1], [(1,), 2], [2, 3]], [(2, 1, 3)])]]


al3c = [[120, ([[1, 2], [1, 3]], [(1,), (2, 3)])]]
         
al3 = al3b + al3c

"""
#COMMENT: al3new \neq al3b + al3c but the difference. 
#Of course, the difference is closed with respect to the 
#differential dConvlc

# the difference:
diff = al3new + mult(-1,al3b+al3c); diff = Simplify(diff)
 
diff = [[-20, ([[(1,), 1], [1, 2], [(1,), 3]], [(1, 2, 3)])],
 [-20, ([[1, (1,)], [(1,), 2], [(1,), 3]], [(1, 2, 3)])],
 [40, ([[(1,), 1], [(1,), 2], [(1,), 3]], [(2, 1, 3)])],
 [20, ([[(1,), 1], [(1,), 2], [2, 3]], [(2, 1, 3)])]]
 
al3new was NOT used to find al4c and al4b.
We used al3b + al3c!  
"""

"""
This piece of code can be used to find arity 4 part: 

CB = loadConvBul(); CC = loadConvCirc()
CB4 = CB[3]; CC4 = CC[3]
del CB; del CC # to go easy on the memory...


Tup4b = tuple(mult(DM, dConvlc([[1,v]])) for v in CB4) 

RHS4b = mult(-1, lieConvlc(al2b, al3b))

Sys4b = System(Tup4b +(RHS4b,)) 

Aug4b = Matrix(Sys4b.Mtx()); Aug4b = Aug4b.T

Res4b = Solve(Aug4b) # Solving the linear system -DM* d(al4b) = - [al2b,al3b]

al4b = toLC(Res4b[1][:],CB4) # it is written explicitly below

# Now we need to store the basis of the ker(dConvlc) \cap span(CB4)

print('I will now form Arb4b and store it in Arb4bfile')
 
Null4b = Res4b[2]; n = Null4b.shape[1] # the number of columns

Arb4b = tuple(toLC(Null4b.col(i)[:],CB4) for i in range(n))
pickle.dump(Arb4b,open('Arb4bfile','wb'))

# now al4c:

Tup4c = tuple(mult(DM, dConvlc([[1,v]])) for v in CC4) 

RHS4c = mult(-1, Simplify(lieConvlc(al2b, al3c) + lieConvlc(al2c, al3b)))

Sys4c = System(Tup4c +(RHS4c,)) 

Aug4c = Matrix(Sys4c.Mtx()); Aug4c = Aug4c.T

Res4c = Solve(Aug4c) # Solving the linear system -DM* d(al4c) = - [al2b,al3c] - [al2c,al3b] 

al4c = toLC(Res4c[1][:],CC4) #it is written explicitly below

# Now we need to store the basis of the ker(dConvlc) \cap span(CC4)

print('I will now form Arb4c and store it in Arb4cfile')
 
Null4c = Res4c[2]; n = Null4c.shape[1] # the number of columns

Arb4c = tuple(toLC(Null4c.col(i)[:],CC4) for i in range(n))
pickle.dump(Arb4c,open('Arb4cfile','wb'))


END for arity 4.
"""
al4c = [[-80, ([[1, 2], [1, 3], [1, 4]], [(1,), (2, 3, 4)])],
 [-20, ([[1, 2], [2, 3], [1, 4]], [(1,), (2, 3, 4)])],
 [-40, ([[1, 2], [1, 3], [1, 4]], [(1,), (3, 2, 4)])],
 [-20, ([[1, 2], [1, 3], [3, 4]], [(1,), (3, 2, 4)])],
 [10, ([[1, 2], [2, 3], [1, 4]], [(1, 2), (3, 4)])],
 [10, ([[1, 2], [2, 3], [1, 4]], [(1, 3), (2, 4)])],
 [10, ([[1, 2], [1, 3], [3, 4]], [(1, 3), (2, 4)])],
 [10, ([[1, 2], [2, 3], [3, 4]], [(1, 3), (2, 4)])],
 [10, ([[1, 2], [1, 3], [3, 4]], [(1, 4), (2, 3)])],
 [10, ([[1, 2], [2, 3], [3, 4]], [(1, 4), (2, 3)])]]
    
 
al4b = [[-60, ([[(1,), 1], [(1,), 2], [(1,), 3], [(1,), 4]], [(1, 2, 3, 4)])],
 [-25, ([[(1,), 1], [1, 2], [(1,), 3], [(1,), 4]], [(1, 2, 3, 4)])],
 [-15, ([[(1,), 1], [(1,), 2], [2, 3], [(1,), 4]], [(1, 2, 3, 4)])],
 [-20, ([[(1,), 1], [1, 2], [1, 3], [(1,), 4]], [(1, 2, 3, 4)])],
 [-10, ([[1, (1,)], [(1,), 2], [(1,), 3], [1, 4]], [(1, 2, 3, 4)])],
 [-5, ([[(1,), 1], [1, 2], [2, 3], [(1,), 4]], [(1, 2, 3, 4)])],
 [-10, ([[1, (1,)], [(1,), 2], [(1,), 3], [(1,), 4]], [(1, 2, 3, 4)])],
 [-5, ([[1, (1,)], [(1,), 2], [2, 3], [(1,), 4]], [(1, 2, 3, 4)])],
 [-30, ([[(1,), 1], [(1,), 2], [(1,), 3], [(1,), 4]], [(1, 3, 2, 4)])],
 [-20, ([[(1,), 1], [1, 2], [(1,), 3], [(1,), 4]], [(1, 3, 2, 4)])],
 [-5, ([[(1,), 1], [(1,), 2], [2, 3], [(1,), 4]], [(1, 3, 2, 4)])],
 [-20, ([[(1,), 1], [1, 2], [1, 3], [(1,), 4]], [(1, 3, 2, 4)])],
 [-10, ([[1, (1,)], [(1,), 2], [(1,), 3], [1, 4]], [(1, 3, 2, 4)])],
 [-5, ([[(1,), 1], [1, 2], [2, 3], [(1,), 4]], [(1, 3, 2, 4)])],
 [-20, ([[(1,), 1], [(1,), 2], [(1,), 3], [3, 4]], [(1, 3, 2, 4)])],
 [-15, ([[(1,), 1], [1, 2], [(1,), 3], [3, 4]], [(1, 3, 2, 4)])],
 [-10, ([[1, 2], [1, (1,)], [(1,), 3], [(1,), 4]], [(1, 3, 2, 4)])],
 [-10, ([[1, (1,)], [(1,), 2], [(1,), 3], [(1,), 4]], [(1, 3, 2, 4)])],
 [-5, ([[1, (1,)], [(1,), 2], [2, 3], [(1,), 4]], [(1, 3, 2, 4)])],
 [-10, ([[1, (1,)], [(1,), 2], [(1,), 3], [3, 4]], [(1, 3, 2, 4)])],
 [5, ([[(1,), 1], [1, 2], [(1,), 3], [(1,), 4]], [(2, 1, 3, 4)])],
 [10, ([[(1,), 1], [(1,), 2], [2, 3], [(1,), 4]], [(2, 1, 3, 4)])],
 [30, ([[(1,), 1], [(1,), 2], [(1,), 3], [(1,), 4]], [(2, 3, 1, 4)])],
 [10, ([[(1,), 1], [1, 2], [(1,), 3], [(1,), 4]], [(2, 3, 1, 4)])],
 [20, ([[(1,), 1], [(1,), 2], [2, 3], [(1,), 4]], [(2, 3, 1, 4)])],
 [20, ([[(1,), 1], [(1,), 2], [(1,), 3], [3, 4]], [(2, 3, 1, 4)])],
 [5, ([[(1,), 1], [1, 2], [(1,), 3], [3, 4]], [(2, 3, 1, 4)])],
 [10, ([[(1,), 1], [(1,), 2], [2, 3], [2, 4]], [(2, 3, 1, 4)])],
 [10, ([[(1,), 1], [(1,), 2], [2, 3], [3, 4]], [(2, 3, 1, 4)])]]
  

"""
#LOOKING FOR TERMS IN ARITY 5 (WITH 1 NEUTRAL VERTEX): 
#
# DM * d(al5b) + [al2b, al4b] + al3b \bullet al3b = 0  
    
#Let us load CB 
 
CB = loadConvBul()
CB5 = CB[4]

del CB # to go easy on the memory...

#Loading the basis of the ker(dConvlc) \cap CB(arity = 4)  
FO = open('Arb4bfile','rb'); Arb4b = pickle.load(FO)

# Now we apply [al2b, ]

print('Computing More')
More = tuple(lieConvlc(al2b,w) for w in Arb4b)

print('Computing RHS')
# This is how we get the RHS:
RHS = Simplify(lieConvlc(al2b, al4b) + pLielc(al3b, al3b))
# Then it was pickled.

#FO = open('RHSfile','rb'); RHS = pickle.load(FO) # Now we have the RHS

print('Computing Btup')
Btup = tuple(mult(-240, dConvlc([[1,t]])) for t in CB5)
# This is how Btup was formed. Then we pickled it into Btupfile

#FO = open('Btupfile','rb')
#Btup = pickle.load(FO) # Now we have the Btup
    
print('Forming the system')
Sys = System(More + Btup +(RHS,))
# We obtained Sys then we pickled it into Sysfile

# To get Sys back:
#FO = open('Sysfile','rb'); Sys = pickle.load(FO) # Now we have the System Sys

LAug = Sys.Mtx()
# Then it was pickled into LAugfile.

#FO = open('LAugfile','rb'); LAug = pickle.load(FO) # Now we have the Augmented matrix in the primitive form

rhs = LAug[len(LAug)-1]
           
#Indices of equations for which the rhs is non-zero:
    
nztup = tuple(i for i in range(len(LAug[0])) if rhs[i] != 0)

#Indices of equations for which the rhs is zero:
      
ztup = tuple(i for i in range(len(LAug[0])) if rhs[i] == 0)

Lmain = [[v[i] for i in nztup] for v in LAug]

print('Computing the sympy matrix AugMain')

AugMain = Matrix(Lmain); AugMain = AugMain.T

#It took more than 2 hours to solve it. 
#The system was consistent.

print('I will now try to find the solution set!')

Main = Solve(AugMain)

#A solution of the inhomogeneous system is pickled to 
#Xfile. The matrix whose columns span the null space is 
#pickled into MNullfile

# Let us try to take care of the NEXT LAYER:
    
# First, unpickle:
    
FO = open('Xfile','rb'); X = pickle.load(FO)

FO = open('MNullfile','rb'); MNull = pickle.load(FO)

Lnext = [[v[i] for i in ztup] for v in LAug[0:len(LAug)-1]]
# ztup is used from the previous step.
         
print('I start to form the augmented matrix of the next layer!')
    
Cf = Matrix(Lnext)
Cf = Cf.T

Y = (-1)*(Cf*X) # this is the rhs of the next layer

Aug = Cf*MNull

Aug = Aug.col_insert(Aug.shape[1],Y) # we need to append the rhs

print('The augmented matrix has the shape', Aug.shape)

#
# SolNext = Solve(Aug) # This line worked for a while
# XX = SolNext[1]; MMNull = SolNext[2]
# pickle.dump(XX, open('XXfile','wb')); pickle.dump(MMNull, open('MMNullfile','wb'))
#
# The above piece of code worked for 2 hours or so. 
# A solution was pickled in XXfile and the matrix for the Null space
# was pickled in MMNullfile


FO = open('XXfile','rb'); XX = pickle.load(FO)

FO = open('MMNullfile','rb'); MMNull = pickle.load(FO)

#Note that X + MNull*XX is a solution of the original system
#and the column space of MNull*MMNull is the solution space of 
#the corresponding homogeneous system. 

print('I will now produce solutions of the original system (arity 5, two neutral vertices)')

YY = X + (MNull*XX)

Nullb = MNull*MMNull 

print('Now I will test')

CL = LAug[0:len(LAug)-1]
          
C = Matrix(CL); C = C.T

w = toCol(rhs)

print('Preparing test and test1')

diff = (C * YY) - w
 
test = NZlist(diff[:])

zero = C*Nullb.col(26)

test1 = NZlist(zero[:])

# The test went fine.

#A solution of the original system was pickled into YYfile.
#The solution of the corresponding homogeneous system is
#the column space of the sympy matrix Nullb. The latter was pickled into Nullbfile.

# Now converting solutions to the actual vectors

FO = open('YYfile','rb'); YY = pickle.load(FO)

FO = open('Nullbfile','rb'); Nullb = pickle.load(FO)

FO = open('Arb4bfile','rb'); Arb4b = pickle.load(FO)

print('I will now compute al4bMore and al5b')

al4bMore = []

for i in range(len(Arb4b)):
    c = YY[:][i]
    if c != 0:
        al4bMore = al4bMore + mult(-c, Arb4b[i])

al4bMore = Simplify(al4bMore)
# al4bMore was pickled into al4bMorefile
        
full = len(YY[:])
L5 = YY[:][len(Arb4b):full]

CB = loadConvBul(); CB5 = CB[4]

al5b = toLC(L5, CB5)
# al5b was pickled into al5bfile

print('Now I will start the test!')

test = mult(240, dConvlc(al5b)) + lieConvlc(al2b, al4b+al4bMore) + pLielc(al3b,al3b)

test = Simplify(test)
# Hurrah!!! This test worked test == []!!!

# NOW WE WILL SOLVE THE SYSTEM INVOLVONG TERMS WITH 1 NEUTRAL VERTEX:
# - d(al5c) - [al2b, al4cMore] =  [al2b,al4c] + [al2c, al4b + al4bMore] + [al3c,al3b]  
#
#
   
FO = open('al4bMorefile','rb'); al4bMore = pickle.load(FO)
FO = open('al5bfile','rb'); al5b = pickle.load(FO)

CC  = loadConvCirc()
CC5= CC[4]
CC4= CC[3]

del CC

Ctup = tuple(mult(-240,dConvlc([[1,t]])) for t in CC5) 

Sys = System(Ctup)

RHS = lieConvlc(al2c,al4b+al4bMore) + lieConvlc(al2b, al4c)+ lieConvlc(al3c,al3b)

RHS = Simplify(RHS)
# Sys.InSpan(RHS)
# My computer worked for 5 hours and showed that RHS is NOT
# in the span of Ctup! This proves that al2c + al2b + al3c + al3b + al4c + al4b + al4bMore
# is NOT the truncation of a genuine MC element. 
# More vectors should be added to al4c. Let us find them!
    
CC  = loadConvCirc()
CC5= CC[4]
CC4= CC[3]
del CC

FO = open('Arb4cfile','rb'); Arb4c = pickle.load(FO)
MoreC = tuple(mult(-1,lieConvlc(al2b,w)) for w in Arb4c)
# (-1)* because we need this term in the LHS
#MoreC was pickled to MoreCfile

FO = open('Arb4bfile','rb'); Arb4b = pickle.load(FO)

SysBrack = System(tuple(lieConvlc(al2b,v) for v in Arb4b))
# The command SysBrack.LinIndep() gave True.
# So we cannot add more terms to al4bMore
# For the next linear system we need only MoreC and Ctup


# Back to the system with 1 neutral vertex:
   
FO = open('al4bMorefile','rb'); al4bMore = pickle.load(FO)

FO = open('MoreCfile','rb'); MoreC = pickle.load(FO)

CC  = loadConvCirc()
CC5= CC[4]
CC4= CC[3]
del CC

Ctup = tuple(mult(-240,dConvlc([[1,t]])) for t in CC5) 

RHSc = lieConvlc(al2c,al4b+al4bMore) + lieConvlc(al2b, al4c)+ lieConvlc(al3c,al3b)
RHSc = Simplify(RHSc)

Sys = System(MoreC+Ctup+(RHSc,))

LAugc = Sys.Mtx()
#LAugc was pickled into LAugcfile 

#Now let us work with the huge system corresponding to LAugc:
    
FO = open('LAugcfile','rb'); LAugc = pickle.load(FO)

rhs = LAugc[len(LAugc)-1] # This is the last entry

nztup = tuple(i for i in range(len(rhs)) if rhs[i] != 0)

L1 = [[v[i] for i in nztup] for v in LAugc]

M1 = Matrix(L1); M1 = M1.T

print('I will now try to solve the "inhomogeneous layer"!')
Sol = Solve(M1) # worked for 1 hour (approx.)
# The system is consistent!
# The results were pickled: Xc to Xcfile, MNullc to MNullcfile


#Now let us take care of the next layer: 
    
FO = open('LAugcfile','rb'); LAugc = pickle.load(FO)

rhs = LAugc[len(LAugc)-1] # This is the last entry
ztup = tuple(i for i in range(len(rhs)) if rhs[i] == 0)

    
FO = open('Xcfile','rb'); Xc = pickle.load(FO)
FO = open('MNullcfile','rb'); MNullc = pickle.load(FO)

Lnext = [[v[i] for i in ztup] for v in LAugc[0:len(LAugc)-1]]
# I do not need the last entry of LAugc

print('I start to form the augmented matrix of the next layer!')
    
Cf = Matrix(Lnext)
Cf = Cf.T

Y = (-1)*(Cf*Xc) # this is the rhs of the next layer

AugC = Cf*MNullc

AugC = AugC.col_insert(AugC.shape[1],Y) # we need to append the rhs
# all the multiplications took less than 15 min!

print('The augmented matrix has the shape', AugC.shape)

pickle.dump(AugC,open('AugCfile','wb'))
# AugC is pickled to AugCfile 

#Now let us find a solution for the next layer:

FO = open('AugCfile','rb'); AugC = pickle.load(FO)

Sol = Solve(AugC) # Solved in minutes!!!
# A particular solution XXc was pickled into XXcfile
# The matrix (whose columns span the null space) MMNullc is pickled into MMNullcfile

# Let us now find a solution of the original system (with 1 neutral vertex):

FO = open('Xcfile','rb'); Xc = pickle.load(FO)
FO = open('XXcfile','rb'); XXc = pickle.load(FO)
FO = open('MNullcfile','rb'); MNullc = pickle.load(FO)
FO = open('MMNullcfile','rb'); MMNullc = pickle.load(FO)

#Note that Xc + MNullc*XXc is a solution of the original system
#and the column space of MNullc*MMNullc is the solution space of 
#the corresponding homogeneous system. 

print('I will now produce solutions of the original system')

YYc = Xc + (MNullc*XXc)

Nullc = MNullc*MMNullc 

print('Now I will test')

FO = open('LAugcfile','rb'); LAugc = pickle.load(FO)

CL = LAugc[0:len(LAugc)-1]
          
C = Matrix(CL)
C = C.T

rhs = LAugc[len(LAugc)-1] # This is the last entry
w = toCol(rhs)

print('Preparing test and test1')

diff = (C * YYc) - w
 
test = NZlist(diff[:])

zero = C*Nullc

test1 = NZlist(zero[:])
# The test worked!
#YYc is pickled into YYcfile

# Now converting the solution to the actual vectors
CC  = loadConvCirc()
CC5= CC[4]
CC4= CC[3]
del CC

FO = open('YYcfile','rb'); YYc = pickle.load(FO)
FO = open('al4bMorefile','rb'); al4bMore = pickle.load(FO)
FO = open('Arb4cfile','rb'); Arb4c = pickle.load(FO)

print('I will now compute al4cMore and al5c')

al4cMore = []

for i in range(len(Arb4c)):
    c = YYc[:][i]
    if c != 0:
        al4cMore = al4cMore + mult(c, Arb4c[i])

al4cMore = Simplify(al4cMore)
        
full = len(YYc[:])
L5 = YYc[:][len(Arb4c):full]

al5c = toLC(L5, CC5)


print('Now I will start the test!')
test = mult(240, dConvlc(al5c))+lieConvlc(al2b, al4c+al4cMore)+lieConvlc(al2c, al4b+al4bMore)+lieConvlc(al3c,al3b)

test = Simplify(test)
# Hurrah!!! It worked test == []!!!

"""
# WE WILL NOW LOAD ALL THE RESULTS:
# (THE COMMENTED LINES BELOW CAN BE USED FOR TESTING)
#

FO = open('al4cMorefile','rb'); al4cMore = pickle.load(FO)
FO = open('al4bMorefile','rb'); al4bMore = pickle.load(FO)
FO = open('al5cfile','rb'); al5c = pickle.load(FO)
FO = open('al5bfile','rb'); al5b = pickle.load(FO)


# al2 = al2c + al2b # al2 was introduced above.
al3 = al3c + al3b

al4 = al4c+ al4cMore + al4b + al4bMore


al4 = Simplify(al4)
# al4 has 85 terms.
# Note that without al4cMore al2+ al3+ al4 cannot be extended!
# It was tested. The test took 5 hours.  
 
al5 = al5c + al5b # No need to Simplify
# al5 has 1173 terms

#This is how we tested it:
#test = mult(240, dConvlc(al5)) + lieConvlc(al2,al4) + pLielc(al3,al3) 
#test = Simplify(test) # We got test == []

#
# ADDITIONAL TESTING: 
#

"""
# TESTING d^2 = 0
#v is a tensor monomial of Conv( , ). 
#sq(v) computes d(d(v)).
#For testing
def sq(v):
    return dConvlc(dConv(v))

CB = loadConvBul()

test = [sq(v) == [] for v in CB[1]] + [sq(v) == [] for v in CB[2]] 
    
test = test + [sq(v) == [] for v in CB[3]]

# test = test + [sq(v) == [] for v in CB[4]] this line took several minutes.
# False in test returned False. So all good!


CC = loadConvCirc()
test = [sq(v) == [] for v in CC[1]] + [sq(v) == [] for v in CC[2]] 
test = test + [sq(v) == [] for v in CC[3]]
test = test + [sq(v) == [] for v in CC[4]] # this lines takes minutes

#False in test returned False. So all good!


# TESTING VIA THE LEIBNIZ RULE: 
    

#v,w are basis vectors in
#Leib(v,w) returns the (simplified) expression 
#[d v, w] + (-1)^{deg(v)}[v, d w] - d([v,w]) 
#For testing
def Leib(v,w):
    t1 = lieConvlc(dConv(v), [[1,w]])
    t2 = mult((-1)**(len(v[0]) + degG(v[1])), lieConvlc([[1,v]], dConv(w)))
    t3 = mult(-1,dConvlc(lieConv(v,w)))
    return Simplify(t1+t2+t3)

CB = loadConvBul()
CC = loadConvBul()

test = []

for t in prod(CB[1], CB[2]):
    test.append(Leib(t[0],t[1]) == [])
    

for t in prod(CB[2], CB[1]):
    test.append(Leib(t[0],t[1]) == [])
    

for t in prod(CB[2], CB[2]):
    test.append(Leib(t[0],t[1]) == [])
    
for t in prod(CC[2], CB[2]):
    test.append(Leib(t[0],t[1]) == [])
    
for t in prod(CC[2], CC[2]):
    test.append(Leib(t[0],t[1]) == [])
    
# It took more than an hour to form the list test. 
# test has 208 entries. 'False in test' returned False. So all good!     
"""

#TESTING VIA THE JACOBI IDENTITY:     
"""
#u,v, w are basis vectors in Conv( , ). 
#Jac(u,v,w) returns the simplified expression 
#[[u,v],w] + (-1)^{deg(u)(deg(v) + deg(w))} [[v,w],u] + (-1)^{deg(w)(deg(u) + deg(v))} [[w,u],v]
#For testing        
def Jac(u,v,w):
    t1 = lieConvlc(lieConv(u,v),[[1,w]])
    exp2 = (len(u[0]) + degG(u[1])) *(len(v[0]) + degG(v[1]) + len(w[0]) + degG(w[1]))
    t2 = mult((-1)**exp2, lieConvlc(lieConv(v,w),[[1,u]]))
    exp3 = (len(w[0]) + degG(w[1])) *(len(u[0]) + degG(u[1]) + len(v[0]) + degG(v[1]))
    t3 = mult((-1)**exp3, lieConvlc(lieConv(w,u),[[1,v]]))
    return Simplify(t1+t2+t3)
    

CB = loadConvBul()
CC = loadConvBul()

test = []
# the loop below will work for a while...
k = 1
for t in prod(CB[1],CC[2],CB[2]):
    print(k); k = k+1 # to see how we are doing
    test.append(Jac(t[0],t[1],t[2])==[])

# The above lines worked less than an hour. 
# 'False in test' returned False. So all good!
""" 



  

    


