"""
Written by Vasily Dolgushev and Geoffrey Schneider.
Tools for working with linear combinations and graphs.
Plus a function which generates partitions. 

Linear combinations are given as lists of "terms".
Each term is a list with two entries.
The first entry is the coefficient and the second entry is
an object which represents a vector in a basis.

For example, if L and LL are two lists representing linear combinations,
then L + LL is the list representing their sum. Note that it is not simplified.
See the function "Simplify" below.

Finally there is a function which converts an n-times-1 matrix 
(of integers) to the corresponding list of length n
"""



import itertools
import time
import math

from sympy import Matrix

from math import factorial as fact

from itertools import combinations_with_replacement as cr

from itertools import combinations as comb
# comb(seq, r) is the generator for 'seq choose r'

from copy import deepcopy as deepcopy

"""
binomC(n,k) returns `n choose k'
It is just for testing.
"""
#Tested
def binomC(n,k):
    return fact(n)//(fact(k)*fact(n-k))


"""
FUNCTIONS FOR WORKING WITH LINEAR COMBINATIONS:
"""



"""
Terms returns the tuple of "terms" of a linear combination vec.
We need "Terms" for "Simplify".
Note that a list cannot be a member in a set :-(
Note that there are no duplicates in Terms(vec).
"""

# Tested
def Terms(vec):
    out = ()
    for el in vec:
        if el[1] not in out:
            out = out + (deepcopy(el[1]),)

    return out
    

"""
"Simplify" simplifies a linear combination vec
"""

# Tested
def Simplify(vec):
    out = []
    for ent in Terms(vec):
        c = 0
        for el in vec:
            if el[1] == ent:
                c = c + el[0]
        if c != 0:
            out.append([c,ent])

    return out

"""
Multiplying the linear combination x by the number k:
"""

#Tested
def mult(k,x):
    xx = deepcopy(x)
    return [[k*el[0], el[1]] for el in xx]

"""
bsum returns the simplified sum of L and LL:
"""

#Tested
def bsum(L, LL):
    return Simplify(deepcopy(L)+deepcopy(LL))

"""
lsum returns the (simplified) sum of the linear combinations in the list L
We need lsum for bilinExt
"""

#Tested
def lsum(L):
    newl = L[0]
    for k in range(1,len(L)):
        newl = newl + deepcopy(L[k])
    return Simplify(newl)


"""
bfunc is a function taking a basis vector to a linear combination. 
linExt(bfunc, x) gives us a linear extension of bfunc.
Of course, the output of linExt is a linear combination.
The function "Simplify" is used!

Note that the output if bfunc must be a linear combination!

"""
#Tested
def linExt(bfunc,x):
    if x ==[]:
        return []
    l = [mult(k[0],bfunc(k[1])) for k in x]
    return lsum(l)
    
         
             
"""
bfunc a function taking
basis vectors to arbitrary linear combinations of basis vectors
"bilinExt" gives us the bilinear externsion of bfunc.
Here x and y are linear combinations.
The outout of bilinExt(bfunc,x,y) is of course a linear combination.
Note: The simplification IS USED!
Note that B(0,?) = B(?,0) = 0 for any bilinear function B.
"""

#Tested
def bilinExt(bfunc, x, y):
    if x == [] or y == []:
        return []
    l = []
    for k in x:
        for j in y:
            l.append(mult(k[0]*j[0],bfunc(k[1],j[1])))
    return lsum(l)

"""
toVect converts a simplified (!) linear combination with elements
in the basis B into the corresponding coordinate vector.
Note: x must be simplified and all terms of x must be multiples of 
vectors from B.
"""
    
# Tested    
def toVect(x, B):
    terms = tuple(el[1] for el in x)
    out = [0]*len(B)
    for e in terms:
        out[B.index(e)] = out[B.index(e)] + x[terms.index(e)][0]
    return out
    
        
"""
toLC converts a coordinate vector v (with respect to the basis B) into
the corresponding linear combination.
Note: v and B must have the same length!
"""

# Tested
def toLC(v,B):
    ind = tuple(i for i in range(len(v)) if v[i] != 0)
    return [[v[i],deepcopy(B[i])] for i in ind]

"""
toLCB converts a coordinate vector v (with respect to the 
basis B) to the corresponding linear combination. 
Elements of B are linear combinations. (Not individual terms!)
Note: v and B must have the same length!
"""
#Tested
def toLCB(v,B):
    out = []
    for i in range(len(v)):
        if v[i] != 0:
            out = out + mult(v[i], deepcopy(B[i]))

    return Simplify(out)



"""
ELEMENTARY FUNCTIONS FOR WORKING WITH GRAPHS:
"""


"""
List L and LL have the same set of elements.
Sign(L,LL) returns the sign factor of
the permutation in S_{len(L)} which
converts L to LL.
The function CAN also work for tuples.
"""

#Tested
def Sign(L,LL):
    s = 1
    for c in comb(tuple(i for i in range(0, len(L))), 2):
        if LL.index(L[c[0]]) > LL.index(L[c[1]]):
            s = s*(-1)

    return s



"""
Part(n,k) generates partitions of n of length k. 
Note that the output is a generator!
"""
# Tested
def Part(n,k):
    for t in cr(range(n,0,-1),k):
        if sum(t) == n:
            yield t


    
"""
Perm(n) generates all permutation in S_n (as tuples).
Notes that the output is a generator!
"""

#Tested
def Perm(n):
	return itertools.permutations(range(1,n+1))



"""
inv(s) returns the inverse of a permutation s:
"""
#Tested
def inv(s):
    n = len(s)
    return tuple(s.index(a)+1 for a in range(1,n+1))

 

"""
Vert returns the tuple of vertices in
the order they appear:
"""

#Tested
def Vert(graf):
    out = tuple()
    for ed in graf:
        for v in ed:
            if v not in out:
                out = out + (v,)

    return out

        

"""
NumVert gives the number of vertices in a graph
"""

#Tested
def NumVert(graf):
    return len(Vert(graf))

"""
Val gives the valency of vertex i in a graph G:
"""
    
#Tested
def Val(G,i):
    v=0
    for ed in G:
        if i in ed:
            v = v+1
    return v


"""
NumIn(T,i) returns the number of edges which originate from 
vertex i. We assume that trees "grow up" 
"""

#Tested
def NumIn(T, i):
    t = tuple(ed for ed in T if ed[0]==i)
    return len(t)



#COMMENT: For an n-times-1 sympy matrix M, M[:] is the list of 
# of the corresponding entries of M.

    
"""
toCol converts a numerical list (of length k) into 
the corresponding k-times-1 matrix. 
"""
#Tested
def toCol(x):
    v = [[a] for a in x]
    return Matrix(v)


"""
Null(C) returns the basis of the nullspace of the sympy matrix C
The output of Null is a list. Each entry of out is 
the list representing the corresponding vector.  
"""
#Tested
def Null(C):
    return [v[:] for v in C.nullspace()]
    


"""
The output of Solve is a tuple (Consist, x_0, NullMtx).
Consist is True if the system is consistent.
x_0 is a solution of the system (as the sympy col matrix) 
and NullMtx is the sympy matrix whose columns form a basis of 
the null space for the coefficient matrix.  
If Consist is False then the function returns (False,(),())
""" 

def Solve(M):
    nvar = M.shape[1]-1
    (S,piv) = M.rref()
    if nvar in piv:
        print('Your system is inconsistent')
        return (False,(),())
    # piv is the list of pivotal entries
    # S is the row reduced augemented matrix
    rhs = S.col(nvar)[:]
    out = [0]*nvar
    for j in piv:
        out[j] = out[j] + rhs[piv.index(j)]
    C = deepcopy(S)
    C.col_del(nvar)# we get the coefficient matrix
    NullM = Matrix(Null(C))
    NullM = NullM.T
    return (True, toCol(out), NullM)
    


"""
NZlist(L) returns the tuple of all non-zero entries of the nested 
numerical list L (without repetitions). 
"""
#Tested
def NZlist(L):
    out = ()
    for a in L:
        if a != 0 and a not in out:
            out = out + (a,)
                
    return out
            
# We needed the stuff BELOW only for testing:

G = [[1,2],[1,3]]

Gtup = ((1,2), (1,3))

L = []
for a in range(1, 5):
    L.append(tuple(i for i in range(1,a+1)))
#Now L is a basis for testing.
    
A = Matrix([
[-1],
[ 3],
[ 2],
[ 1]])

def concat(t,tt):
    return [[1,t+tt]]


def plus(t):
    return [[1,t+('love',)]]
             
c1 = [[3, (1,)], [-2, (1, 2, 3)], [5, (1, 2, 3)]]

v = [3, 0, -2, 0, 0, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

c = [[3, (1,)], [-2, (1, 2, 3)], [5, (1, 2, 3, 4, 5, 6)]]
             
    
             
        


    


