"""
This program was written jointly by Vasily Dolgushev and Geoffrey Schneider

This module contains functions for converting "computer-presented" 
(c-presented) brace trees to "human presented" (h-presented) brace trees. 

For h-presented brace trees the labeled vertices have their usual labels 
and the neutral vertices are labeled by tuples of length 1: (1,), (2,), ...
For example the cup-product in h-presentation is 
    
hCup = [ [(1), 1], [(1), 2] ] 

the c-presentation for this brace tree is 

Cup = [ 1, [[1,2],[1,3]] ]

h-presentation has the obvious advantage: usual BT-trees (i.e. brace 
trees without neutral vertices are automatically h-presented brace trees). 
"""




# coding: utf-8
#import pickle


from itertools import combinations_with_replacement as cr
from itertools import permutations as perm
from itertools import combinations as comb
from copy import deepcopy as deepcopy


#import time
#import math

from LinCombGraphs import linExt as linExt
from LinCombGraphs import NumVert as NumVert
from LinCombGraphs import NumIn as NumIn



"""
VertBr returns the tuple of vertices of an h-presented
brace tree. The vertices appear in the standard order which 
comes from the planar structure. 
No repetitions in this tuple!
"""

#Tested
def VertBr(T):
    t = ()
    for ed in T:
        for a in ed:
            if a not in t:
                t = t + (a,)
    
    return t


"""
countNV returns the number of neutral vertices 
of an h-presented tree T
"""
#Tested
def countNV(T):
    t = tuple(a for a in VertBr(T) if type(a) == tuple)
    return len(t)
    
"""
arity(T) returns the number of labeled vertices of T. 
"""
#Tested
def arity(T):
    return len(T)+1 - countNV(T)



"""
H2Comp returns the c-presented (computer-presented)
brace tree corresponding to an h-presented brace tree T 
"""
#Tested
def H2Comp(T):
    r = countNV(T)
    def NL(a):
        if type(a) == tuple:
            return a[0]
        else:
            return r+a
    
    return [r, [ [NL(ed[0]), NL(ed[1])] for ed in T] ] 
        
"""
Comp2H is the inverse of H2Comp
"""
#Tested
def Comp2H(T):
    r = T[0]
    def NLinv(a):
        if a <= r:
            return (a,)
        else:
            return a-r
    
    return [ [NLinv(ed[0]), NLinv(ed[1])] for ed in T[1] ]


"""
Comp2Hlc is the extension of Comp2H to linear combinations.
"""
#Tested
def Comp2Hlc(L):
    def f(T):
        return [[1,Comp2H(T)]]
    
    return linExt(f, L)

    
"""
getPermBr(T) returns a permutation sig such that T = sig(Tpr),
where Tpr has the standard order of vertices.
"""
#Tested
def getPermBr(T):
    return tuple(v for v in VertBr(T) if type(v) == int)

    
"""
sig is tuple representing an element of S_n and T is a brace 
tree with n labeled vertices.
ActBr(sig,T) returns sig(T).
Here T is an h-presented brace tree.
"""
#Tested
def ActBr(sig, T):
    out = []
    for ed in T:
        if type(ed[0]) == tuple:
            a = ed[0]
        else:
            a = sig[ed[0]-1]
        if type(ed[1]) == tuple:
            b = ed[1]
        else:
            b = sig[ed[1]-1]
        out.append([a,b])

    return out

    
    
"""
T is a brace tree without neutral vertices and with the standard order 
of labels. r is a positive integer < NumVert(T)
giveBr(r, T) generates all braces trees with r neutral vertices 
which can be obtained from T by replacing labeled vertices by neutral vertices. 
Note: for our procedure, we only need r = 1. 
Note that giveBr generates only admissible brace trees. I.e. every 
neutral vertex has at least 2 children. 
"""
#Tested
def giveBr(r,T):
    n = NumVert(T)
    okvert = tuple(i for i in range(1,n+1) if NumIn(T,i) > 1)
    for c in comb(okvert,r):
        TT , cc = [ ], tuple(i for i in range(1, n+1) if i not in c)
        for ed in T:
            if ed[0] in c:
                v = (c.index(ed[0])+1,)
            else:
                v = cc.index(ed[0])+1 # if ed[0] is not in c 
            if ed[1] in c:
                vv = (c.index(ed[1])+1,)
            else:
                vv = cc.index(ed[1])+1 # if ed[1] is not in c
            TT.append([v,vv])
        yield TT



# This is for testing:
TT = [[1,2],[2,3],[2,4],[2,5],[1,6],[6,7],[6,8]]
          

